#include "hpp/hik/slack-solver.hh"
#include <string.h>
using namespace boost::numeric::ublas;

extern "C"
{
#include "fsqp/cfsqpusr.h"
}

void slackobj32(int nparam, int j,double *x,double *fj,void *cd);
void slackgrob32(int nparam, int j,double *x, double *gradfj,void (*dummy)(int,int,double *,double *,void *), void *cd);
void slackcntr32(int nparam, int j, double *x, double *gj, void *cd);
void slackgrcn32(int nparam, int j, double *x, double *gradgj, void (*dummy)(int,int,double *,double *,void *), void *cd);

void slackobj32(int nparam, int j,double *x,double *fj,void *cd)
{
    //std::cout << "start slackobj32" << std::endl;
    ChppHikSlackSolver *solver = (ChppHikSlackSolver *)cd;

    *fj = 0.0;
    unsigned int i,k = nparam-(int)solver->inequalityTaskCount();
    for (i=k; i<(unsigned int)nparam; i++)
        *fj += x[i]*x[i];

    vectorN dq(k);
    for (i=0; i<k; i++)
    {
        dq[i] = x[i];
        //damping
        *fj += x[i]*x[i]*solver->dampingFactor();
    }

    for (i=0;i<solver->equalityTasks().size();i++)
    {

        vectorN val = prod(solver->equalityTasks()[i]->jacobian(),dq) - solver->equalityTasks()[i]->value();
        *fj += inner_prod(val,val);
    }
    
    
            
    //std::cout << "end slackobj32" << std::endl;
    return;
}

void slackgrob32(int nparam, int j,double *x, double *gradfj,void (*dummy)(int,int,double *,double *,void *), void *cd)
{
    //std::cout << "start slackgrob32" << std::endl;
    ChppHikSlackSolver *solver = (ChppHikSlackSolver *)cd;

    unsigned int k = nparam-(int)solver->inequalityTaskCount(),i,h;
    unsigned int szEqT = solver->equalityTasks().size();
            
    vectorN dq(k),gr(k);
    for (i=0; i<k; i++)
        dq(i) = x[i];
    
    for (i=0; i<k; i++)
        gradfj[i] = 2.0*solver->dampingFactor()*x[i] ; // damping

    for (i=k; i<(unsigned int)nparam; i++)
        gradfj[i] = 2.0*x[i];

    for (i=0;i<szEqT;i++)
    {
        vectorN val = prod(solver->equalityTasks()[i]->jacobian(),dq) - solver->equalityTasks()[i]->value();
        noalias(gr) = 2.0*prod(trans(solver->equalityTasks()[i]->jacobian()),val);
        for (h=0; h<k; h++)
            gradfj[h] += gr[h];
    }

    //std::cout << "end slackgrob32" << std::endl;
    return;
}

void slackcntr32(int nparam, int j, double *x, double *gj, void *cd)
{
    //std::cout << "start slackcntr32" << std::endl;
    ChppHikSlackSolver *solver = (ChppHikSlackSolver *)cd;

    solver->valueOfConstraint(j-1,x,*gj);
    //std::cout << "end slackcntr32" << std::endl;
    return;
}

void slackgrcn32(int nparam, int j, double *x, double *gradgj, void (*dummy)(int,int,double *,double *,void *), void *cd)
{
    //std::cout << "start slackgrcn32" << std::endl;
    ChppHikSlackSolver *solver = (ChppHikSlackSolver *)cd;

    solver->gradientOfConstraint(j-1, gradgj);
    //std::cout << "end slackgrcn32" << std::endl;
    return;
}

const std::vector<CjrlLinearConstraint*>& ChppHikSlackSolver::equalityTasks()
{
    return attEqTaskBlocs;
}

double ChppHikSlackSolver::dampingFactor() const
{
   return attDamping;
}

ChppHikSlackSolver::ChppHikSlackSolver(unsigned int i_ParameterNumber)
{
    attVariableCount = attParamCount = i_ParameterNumber;
    attEqTaskSize = attIneqTaskSize = attEqConstSize = attIneqConstSize = 0;

    mode=100;
    iprint=0;
    //iprint=1;
    miter=500;
    bigbnd=1e10;
    eps=1e-5;
    epsneq=0e0;
    udelta=0e0;
    neqn=0; /* Number of nonlinear equality constraints */
    nineqn=0; /* Number of nonlinear inequality constraints */
    ncsrl=ncsrn=nfsr=mesh_pts[0]=0;

    attDefaultLBound.resize(attParamCount);
    attDefaultUBound.resize(attParamCount);

    for (unsigned int i = 0; i<attParamCount;i++)
    {
        attDefaultLBound(i) = -bigbnd;
        attDefaultUBound(i) = bigbnd;
    }

    attLBound.resize(attParamCount);
    attUBound.resize(attParamCount);

    tempdq.resize(attParamCount);
    tempro.resize(attParamCount);

    clearBounds();

    attParams.resize(attParamCount);
}

ChppHikSlackSolver::~ChppHikSlackSolver()
{
}


int ChppHikSlackSolver::solve(vectorN& io_solution, vectorN& outSlackVariables, double damping)
{
    attDamping = damping*damping;
    if (io_solution.size() != attParamCount )
    {
        std::cout << "ChppHikSlackSolver::solve incorrect size for initial point. Aborted." << std::endl;
        return -1000;
    }

    int neq, nineq, inform;
    void *cd = (void *)this;

    nf = 1;
    neq = attEqConstSize;
    nineq = attIneqTaskSize+attIneqConstSize;

    bl=(double *)calloc(attVariableCount,sizeof(double));
    bu=(double *)calloc(attVariableCount,sizeof(double));
    g=(double *)calloc(nineq+neq,sizeof(double));
    lambda=(double *)calloc(nineq+neq+nf+attVariableCount,sizeof(double));
    f=(double *)calloc(nf,sizeof(double));

    memcpy(bl,(&attLBound.data()[0]),attParamCount*sizeof(double));
    memcpy(bu,(&attUBound.data()[0]),attParamCount*sizeof(double));

    for (unsigned int i = attParamCount; i<attVariableCount;i++)
    {
        bl[i] = 0.0;
        bu[i] = bigbnd;
    }

    attFullState = zero_vector<double>(attVariableCount);
    subrange(attFullState,0,attParamCount) = io_solution;

    cfsqp(attVariableCount,nf,nfsr,nineqn,nineq,neqn,neq,ncsrl,ncsrn,mesh_pts,
          mode,iprint,miter,&inform,bigbnd,eps,epsneq,udelta,bl,bu,
          attFullState.data().begin(),
          f,g,lambda,slackobj32,slackcntr32,slackgrob32,slackgrcn32,cd);

    if (f) free(f);
    if (g) free(g);
    if (bl) free(bl);
    if (bu) free(bu);
    if (lambda) free(lambda);

    io_solution = subrange(attFullState,0,attParamCount);
    outSlackVariables = subrange(attFullState,attParamCount,attVariableCount);
    
    return inform;
}

bool ChppHikSlackSolver::valueOfConstraint(unsigned int qpCRank, double *inX, double& outValue)
{
    if(qpCRank >= attIneqTaskSize+attIneqConstSize+attEqConstSize)
    {
        std::cerr << "ChppHikSlackSolver::valueOfConstraint requested rank " <<qpCRank <<" is out of bounds" <<std::endl;
        return false;
    }

    unsigned int crank,wrank;
    std::pair<CjrlLinearConstraint*,unsigned int> cpair;

    outValue = 0.0;

    if (qpCRank < attIneqTaskSize)//Ineq tasks
    {
        crank = qpCRank;
        wrank = attParamCount + crank;
        cpair = attIneqTasks[crank];
        outValue -= inX[wrank];
    }
    else
        if (qpCRank < attIneqTaskSize + attIneqConstSize)//Ineq Constraints
        {
            crank = qpCRank - attIneqTaskSize;
            cpair = attIneqConstraints[crank];
        }
        else
            //Eq Constraints
        {
            crank = qpCRank - attIneqTaskSize - attIneqConstSize;
            cpair = attEqConstraints[crank];
        }


    outValue -= cpair.first->value()(cpair.second);
    memcpy((&tempdq.data()[0]),inX,attParamCount*sizeof(double));
    noalias(tempro) = row(cpair.first->jacobian(), cpair.second);
    outValue += inner_prod(row(cpair.first->jacobian(), cpair.second),tempdq);
    return true;
}

bool ChppHikSlackSolver::gradientOfConstraint(unsigned int qpCRank, double* outVector)
{
    unsigned int i;
    if(qpCRank >= attIneqTaskSize+attIneqConstSize+attEqConstSize)
    {
        std::cerr << "ChppHikSlackSolver::gradientOfConstraint requested rank " <<qpCRank <<" is out of bounds" <<std::endl;
        return false;
    }

    for (i = attParamCount;i<attVariableCount;i++)
        outVector[i] = 0.0;

    unsigned int crank,wrank;
    std::pair<CjrlLinearConstraint*,unsigned int> cpair;

    if (qpCRank < attIneqTaskSize)
    {
        crank = qpCRank;
        wrank = attParamCount + crank;
        cpair = attIneqTasks[crank];
        outVector[wrank] = -1.0;
    }
    else
        if (qpCRank < attIneqTaskSize + attIneqConstSize)
        {
            crank = qpCRank - attIneqTaskSize;
            cpair = attIneqConstraints[crank];
        }
        else
        {
            crank = qpCRank - attIneqTaskSize - attIneqConstSize;
            cpair = attEqConstraints[crank];
        }

    unsigned int offset = cpair.second*attParamCount;

    for (i = 0;i<attParamCount;i++)
        outVector[i] = cpair.first->jacobian().data()[offset+i];


    return true;
}


bool ChppHikSlackSolver::addTasks(CjrlLinearConstraint *i_constraint, bool i_isEquality)
{
    if (i_constraint->jacobian().size2() != attParamCount)
    {
        std::cerr << "ChppHikSlackSolver::addTasks incorrect size of jacobian" <<std::endl;
        return false;
    }
    if (i_isEquality)
    {
        for (unsigned int i=0;i<i_constraint->dimension();i++)
            attEqTasks.push_back(std::make_pair(i_constraint,i));
        attEqTaskSize += i_constraint->dimension();
        attEqTaskBlocs.push_back(i_constraint);
    }
    else
    {
        for (unsigned int i=0;i<i_constraint->dimension();i++)
            attIneqTasks.push_back(std::make_pair(i_constraint,i));
        attIneqTaskSize += i_constraint->dimension();
        attVariableCount +=i_constraint->dimension();
    }
    return true;
}

bool ChppHikSlackSolver::addConstraints(CjrlLinearConstraint *i_constraint, bool i_isEquality)
{
    if (i_constraint->jacobian().size2() != attParamCount)
    {
        std::cerr << "ChppHikSlackSolver::addConstraints incorrect size of jacobian" <<std::endl;
        return false;
    }
    if (i_isEquality)
    {
        for (unsigned int i=0;i<i_constraint->dimension();i++)
            attEqConstraints.push_back(std::make_pair(i_constraint,i));
        attEqConstSize += i_constraint->dimension();
    }
    else
    {
        for (unsigned int i=0;i<i_constraint->dimension();i++)
            attIneqConstraints.push_back(std::make_pair(i_constraint,i));
        attIneqConstSize += i_constraint->dimension();
    }
    return true;
}

unsigned int ChppHikSlackSolver::equalityTaskCount() const
{
    return attEqTaskSize;
}

unsigned int ChppHikSlackSolver::inequalityTaskCount() const
{
    return attIneqTaskSize;
}

unsigned int ChppHikSlackSolver::equalityConstraintCount() const
{
    return attEqConstSize;
}

unsigned int ChppHikSlackSolver::inequalityConstraintCount() const
{
    return attIneqConstSize;
}

void ChppHikSlackSolver::clearTasks()
{
    attEqTasks.clear();
    attEqTaskBlocs.clear();
    attIneqTasks.clear();
    attVariableCount -= attIneqTaskSize;
    attIneqTaskSize = 0;
    attEqTaskSize = 0;

}
void ChppHikSlackSolver::clearConstraints()
{
    attIneqConstraints.clear();
    attEqConstraints.clear();
    attIneqConstSize = 0;
    attEqConstSize = 0;
}

void ChppHikSlackSolver::clearBounds()
{
    unsigned int i;
    for (i=0; i<attParamCount; i++)
    {
        attLBound(i)=attDefaultLBound(i);
        attUBound(i)=attDefaultUBound(i);
    }
}

void ChppHikSlackSolver::upperBound(unsigned int i_rank, double i_bound)
{
    if (i_rank >= attParamCount)
    {
        std::cerr << "ChppHikSlackSolver::upperBound() : rank is too big("
        << i_rank << ")" << std::endl;
        return;
    }
    if (i_bound < attLBound(i_rank))
    {
        std::cerr << "ChppHikSlackSolver::upperBound() : new upper bound("
        << i_bound << ") for " << i_rank
        << "th parameter is smaller than lower bound("
        << attLBound[i_rank] << ")" << std::endl;
        return;
    }
    if (attUBound(i_rank) > i_bound)
        attUBound(i_rank) = i_bound;
}

void ChppHikSlackSolver::lowerBound(unsigned int i_rank, double i_bound)
{
    if (i_rank >= attParamCount)
    {
        std::cerr << "ChppHikSlackSolver::lowerBound() : rank is too big("
        << i_rank << ")" << std::endl;
        return;
    }
    if (i_bound > attUBound(i_rank))
    {
        std::cerr << "ChppHikSlackSolver::lowerBound() : new lower bound("
        << i_bound << ") for " << i_rank
        << "th parameter is bigger than upper bound("
        << attUBound[i_rank] << ")" << std::endl;
        return;
    }
    if (attLBound(i_rank) < i_bound)
        attLBound(i_rank) = i_bound;
}

void ChppHikSlackSolver::defaultUpperBound(unsigned int i_rank, double i_bound)
{
    if (i_rank >= attParamCount)
    {
        std::cerr << "ChppHikSlackSolver::defaultUpperBound() : rank is too big("
        << i_rank << ")" << std::endl;
        return;
    }
    attDefaultUBound[i_rank]  = i_bound;
}

void ChppHikSlackSolver::defaultLowerBound(unsigned int i_rank, double i_bound)
{
    if (i_rank >= attParamCount)
    {
        std::cerr << "ChppHikSlackSolver::defaultLowerBound() : rank is too big("
        << i_rank << ")" << std::endl;
        return;
    }
    attDefaultLBound(i_rank)  = i_bound;
}

double ChppHikSlackSolver::upperBound(unsigned int i_rank) const
{
    if (i_rank >= attParamCount)
    {
        std::cerr << "ChppHikSlackSolver::upperBound() : rank is too big("
        << i_rank << ")" << std::endl;
        return 0;
    }
    return attUBound(i_rank);
}

double ChppHikSlackSolver::lowerBound(unsigned int i_rank) const
{
    if (i_rank >= attParamCount)
    {
        std::cerr << "ChppHikSlackSolver::lowerBound() : rank is too big("
        << i_rank << ")" << std::endl;
        return 0;
    }
    return attLBound(i_rank);
}
